@@ -2,13 +2,9 @@
 #
 #  Net::Server::PSGI - Extensible Perl HTTP PSGI base server
 #
-#  $Id$
+#  Copyright (C) 2011-2022
 #
-#  Copyright (C) 2011-2012
-#
-#    Paul Seamons
-#    paul@seamons.com
-#    http://seamons.com/
+#    Paul Seamons <paul@seamons.com>
 #
 #  This package may be distributed under the terms of either the
 #  GNU General Public License
@@ -21,6 +17,7 @@ package Net::Server::PSGI;
 
 use strict;
 use base qw(Net::Server::HTTP);
+use Scalar::Util qw(blessed);
 
 sub net_server_type { __PACKAGE__ }
 
@@ -54,7 +51,7 @@ sub process_request {
         $self->process_headers;
 
         alarm($self->timeout_idle);
-        my $env = \%ENV;
+        my $env = { %ENV };
         $env->{'psgi.version'}      = [1, 0];
         $env->{'psgi.url_scheme'}   = ($ENV{'HTTPS'} && $ENV{'HTTPS'} eq 'on') ? 'https' : 'http';
         $env->{'psgi.input'}        = $self->{'server'}->{'client'};
@@ -206,7 +203,7 @@ rudimentary, then Net::Server::PSGI may be good for you.
 Net::Server::PSGI takes Net::Server::HTTP one level farther.  It
 begins with base type MultiType defaulting to Net::Server::Fork.  It
 is easy to change it to any of the other Net::Server flavors by
-passing server_type => $other_flavor in the server configurtation.
+passing server_type => $other_flavor in the server configuration.
 The port has also been defaulted to port 80 - but could easily be
 changed to another through the server configuration.  You can also
 very easily add ssl by including, proto=>"ssl" and provide a
